'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _assert = require('assert');

var _assert2 = _interopRequireDefault(_assert);

var _selectn = require('selectn');

var _selectn2 = _interopRequireDefault(_selectn);

var _values = require('graphql/execution/values');

var _graphql = require('graphql');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function costAnalysisMessage(max, actual) {
  return 'The query exceeds the maximum cost of ' + max + '. ' + ('Actual cost is ' + actual);
}

var CostAnalysis = function () {
  function CostAnalysis(context, options) {
    _classCallCheck(this, CostAnalysis);

    (0, _assert2.default)(typeof options.maximumCost === 'number' && options.maximumCost > 0, 'Maximum query cost must be a positive number');

    this.context = context;
    this.cost = 0;
    this.options = options;
    this.multipliers = [];
    this.defaultCost = this.options.defaultCost || 0;

    this.OperationDefinition = {
      enter: this.onOperationDefinitionEnter,
      leave: this.onOperationDefinitionLeave
    };
  }

  _createClass(CostAnalysis, [{
    key: 'onOperationDefinitionEnter',
    value: function onOperationDefinitionEnter(operation) {
      switch (operation.operation) {
        case 'query':
          this.cost += this.computeNodeCost(operation, this.context.getSchema().getQueryType());
          break;
        case 'mutation':
          this.cost += this.computeNodeCost(operation, this.context.getSchema().getMutationType());
          break;
        case 'subscription':
          this.cost += this.computeNodeCost(operation, this.context.getSchema().getSubscriptionType());
          break;
        default:
          throw new Error('Query Cost could not be calculated for operation of type ' + operation.operation);
      }
    }
  }, {
    key: 'onOperationDefinitionLeave',
    value: function onOperationDefinitionLeave() {
      if (this.options.onComplete) {
        this.options.onComplete(this.cost);
      }

      if (this.cost > this.options.maximumCost) {
        return this.context.reportError(this.createError());
      }
    }
  }, {
    key: 'computeCost',
    value: function computeCost(_ref) {
      var multiplier = _ref.multiplier,
          _ref$useMultipliers = _ref.useMultipliers,
          useMultipliers = _ref$useMultipliers === undefined ? true : _ref$useMultipliers,
          _ref$complexity = _ref.complexity,
          complexity = _ref$complexity === undefined ? 1 : _ref$complexity;

      if (complexity > 10 || complexity < 1) {
        this.context.reportError(new _graphql.GraphQLError('The complexity argument must be between 1 and 10'));
        return this.defaultCost;
      }

      if (useMultipliers) {
        if (multiplier) {
          this.multipliers.push(parseInt(multiplier));
        }
        return this.multipliers.reduce(function (acc, multiplier) {
          return acc * multiplier;
        }, complexity);
      }
      return complexity;
    }
  }, {
    key: 'computeCostFromTypeMap',
    value: function computeCostFromTypeMap(node, parentType, fieldArgs) {
      var costObject = this.options.costMap && this.options.costMap[parentType] && this.options.costMap[parentType][node.name.value];

      if (!costObject) {
        return this.defaultCost;
      }

      var useMultipliers = costObject.useMultipliers,
          multiplier = costObject.multiplier,
          complexity = costObject.complexity;

      multiplier = multiplier && (0, _selectn2.default)(multiplier, fieldArgs);

      return this.computeCost({ useMultipliers: useMultipliers, multiplier: multiplier, complexity: complexity });
    }
  }, {
    key: 'computeCostFromDirectives',
    value: function computeCostFromDirectives(directives, fieldArgs) {
      var costDirective = directives.find(function (directive) {
        return directive.name.value === 'cost';
      });
      if (costDirective && costDirective.arguments) {
        // get cost arguments
        var complexityArg = costDirective.arguments.find(function (arg) {
          return arg.name.value === 'complexity';
        });
        var useMultipliersArg = costDirective.arguments && costDirective.arguments.find(function (arg) {
          return arg.name.value === 'useMultipliers';
        });
        var multiplierArg = costDirective.arguments && costDirective.arguments.find(function (arg) {
          return arg.name.value === 'multiplier';
        });

        // get arguments's values
        var _useMultipliers = useMultipliersArg && useMultipliersArg.value && useMultipliersArg.value.kind === 'BooleanValue' ? useMultipliersArg.value.value : true;

        var _multiplier = multiplierArg && multiplierArg.value.value ? parseInt((0, _selectn2.default)(multiplierArg.value.value, fieldArgs)) : undefined;

        var _complexity = complexityArg && complexityArg.value && complexityArg.value.kind === 'IntValue' ? parseInt(complexityArg.value.value) : 1;

        return this.computeCost({ multiplier: _multiplier, useMultipliers: _useMultipliers, complexity: _complexity });
      }
      return this.defaultCost;
    }
  }, {
    key: 'computeNodeCost',
    value: function computeNodeCost(node, typeDef) {
      var _this = this;

      if (!node.selectionSet) {
        return 0;
      }
      var fields = {};
      if (typeDef instanceof _graphql.GraphQLObjectType || typeDef instanceof _graphql.GraphQLInterfaceType) {
        fields = typeDef.getFields();
      }
      return node.selectionSet.selections.reduce(function (total, childNode) {
        var nodeCost = _this.defaultCost;

        // empty array of multipliers if field is at the root of an operation
        if (node.kind === _graphql.Kind.OPERATION_DEFINITION) {
          _this.multipliers = [];
        }

        switch (childNode.kind) {
          case _graphql.Kind.FIELD:
            {
              var field = fields[childNode.name.value];
              // Invalid field, should be caught by other validation rules
              if (!field) {
                break;
              }
              var fieldType = (0, _graphql.getNamedType)(field.type);

              // get field's arguments
              var fieldArgs = {};
              try {
                fieldArgs = (0, _values.getArgumentValues)(field, childNode, _this.options.variables || {});
              } catch (e) {
                _this.context.reportError(e);
              }

              // it the costMap option is set, compute the cost with the costMap provided
              // by the user.
              if (_this.options.costMap && _typeof(_this.options.costMap) === 'object') {
                nodeCost = typeDef && typeDef.name ? _this.computeCostFromTypeMap(childNode, typeDef.name, fieldArgs) : _this.defaultCost;
              } else {
                // Compute cost of current field with its directive
                var costIsComputed = false;
                if (field.astNode && field.astNode.directives) {
                  nodeCost = _this.computeCostFromDirectives(field.astNode.directives, fieldArgs);
                  var costDirective = field.astNode.directives.find(function (directive) {
                    return directive.name.value === 'cost';
                  });
                  if (costDirective && costDirective.arguments) {
                    costIsComputed = true;
                  }
                }
                // if the cost directive is defined on the Type
                // and the nodeCost has not already been computed
                if (fieldType && fieldType.astNode && fieldType.astNode.directives && fieldType instanceof _graphql.GraphQLObjectType && costIsComputed === false) {
                  nodeCost = _this.computeCostFromDirectives(fieldType.astNode.directives, fieldArgs);
                }
              }

              var childCost = 0;
              childCost = _this.computeNodeCost(childNode, fieldType);
              nodeCost += childCost;
              break;
            }
          case _graphql.Kind.FRAGMENT_SPREAD:
            {
              var fragment = _this.context.getFragment(childNode.name.value);
              var fragmentType = fragment && _this.context.getSchema().getType(fragment.typeCondition.name.value);
              nodeCost = fragment ? _this.computeNodeCost(fragment, fragmentType) : _this.defaultCost;
              break;
            }
          case _graphql.Kind.INLINE_FRAGMENT:
            {
              var inlineFragmentType = typeDef;
              if (childNode.typeCondition && childNode.typeCondition.name) {
                // $FlowFixMe: don't know why Flow thinks it could be undefined
                inlineFragmentType = _this.context.getSchema().getType(childNode.typeCondition.name.value);
              }
              nodeCost = childNode ? _this.computeNodeCost(childNode, inlineFragmentType) : _this.defaultCost;
              break;
            }
          default:
            {
              nodeCost = _this.computeNodeCost(childNode, typeDef);
              break;
            }
        }
        return Math.max(nodeCost, 0) + total;
      }, 0);
    }
  }, {
    key: 'createError',
    value: function createError() {
      if (typeof this.options.createError === 'function') {
        return this.options.createError(this.options.maximumCost, this.cost);
      }
      return new _graphql.GraphQLError(costAnalysisMessage(this.options.maximumCost, this.cost));
    }
  }]);

  return CostAnalysis;
}();

exports.default = CostAnalysis;